//#define __DEBUG__
//#define __WARN__

#ifdef ESP32
const int controllerPin = 33;
const int roasterPin = 32;
#else
const int controllerPin = 2; // white
const int roasterPin = 3; // green
#endif


unsigned long timeIntervals[56];
uint8_t bytebuffer[7];

const int preamble = 7000;
const int one_length = 1200;
const int roasterLength = 7;
const int controllerLength = 6;

bool toggle = 0;
double temp = 0.0;
int heat = 0;
int air = 0;

char CorF = 'C';

void setup() {
  Serial.begin(115200);
}


/*double calculateTempJosh() {

    //I really hate this. 
    //It seems to work nicely.. but I feel like there must be a better way than 
    //using a 4th degree polynomial to model this but I sure can't seem to figure it out. 

  double x = ((bytebuffer[0] << 8) + bytebuffer[1]) / 1000.0;
  double y = ((bytebuffer[2] << 8) + bytebuffer[3]) / 1000.0;

#ifdef __DEBUG__
  Serial.print(x);
  Serial.print(',');
  Serial.println(y);
#endif

  float v = 583.1509258523457 + -714.0345395202813 * x + -196.071718077524 * y
         + 413.37964344228334 * x * x + 2238.149675349052 * x * y
         + -4099.91031297056 * y * y + 357.49007607425233 * x * x * x
         + -5001.419602972793 * x * x * y + 8242.08618555862 * x * y * y
         + 247.6124684730026 * y * y * y + -555.8643213534281 * x * x * x * x
         + 3879.431274654493 * x * x * x * y + -6885.682277959339 * x * x * y * y
         + 2868.4191998911865 * x * y * y * y + -1349.1588373011923 * y * y * y * y;

  if (CorF == 'C') v = 5.0f / 9 * (v - 32); // to Celsius
  
  return v;

}*/

uint16_t intx;
uint16_t inty;

double calculateTemp() {

	/*
	Using y data set for temperatures lower than 110C, and x for higher
	Simpler formulas and better resolution
	Computed as Celsius values directly
	*/
	intx = ((bytebuffer[0] << 8) + bytebuffer[1]);
	inty = ((bytebuffer[2] << 8) + bytebuffer[3]);

	double x = 0.001 * intx;
	double y = 0.001 * inty;

#ifdef __DEBUG__
	Serial.print("x,y: ");
	Serial.print(x);
	Serial.print(',');
	Serial.println(y);
#endif

	double v;
	if (intx > 836 || inty > 221)
		v = -224.2 * y * y * y + 385.9 * y * y - 327.1 * y + 171;
	else
		v = -278.33 * x * x * x + 491.944 * x * x - 451.444 * x + 310.668;

	if (CorF == 'F') v = 1.8f * v + 32; // to Fahr; default is C

	return v;

}

bool getMessage(int bytes, int pin) {
  unsigned long pulseDuration = 0;
  int bits = bytes * 8;

  while (pulseDuration < preamble) {  //Wait for it...
    pulseDuration = pulseIn(pin, LOW);
	//Serial.println(pulseDuration);
	//if (pulseDuration == 0) return 0; // pulse one second timeout, abandon, to unlock program flow
  }

  //Serial.print(millis()); 
  //Serial.print(" got preamble ");
  //if (bytes == controllerLength)
	 // Serial.print(" control panel ");
  //else
	 // Serial.print(" roaster ");

  //Serial.println("");
  //return;

  for (int i = 0; i < bits; i++) {  //Read the proper number of bits..
    timeIntervals[i] = pulseIn(pin, LOW);
  }

  for (int i = 0; i < 7; i++) {  //zero that buffer
    bytebuffer[i] = 0;
  }

  for (int i = 0; i < bits; i++) {  //Convert timings to bits
    //Bits are received in LSB order..
    if (timeIntervals[i] > one_length) {  // we received a 1
      bytebuffer[i / 8] |= (1 << (i % 8));
    }
  }

  return 1;

} // end getMessage

bool calculateChecksum(int length) {
  uint8_t sum = 0;
  for (int i = 0; i < (length - 1); i++) {
    sum += bytebuffer[i];
  }
#ifdef __DEBUG__
  Serial.print("sum: ");
  Serial.print(sum, HEX);
  Serial.print(" Checksum Byte: ");
  Serial.println(bytebuffer[length - 1], HEX);
#endif
  return sum == bytebuffer[length - 1];
}

void printBuffer(int bytes) {
  for (int i = 0; i < bytes; i++) {
    Serial.print(bytebuffer[i], HEX);
    Serial.print(',');
  }
  Serial.print("\n");
}

void getControllerMessage() {
#ifdef __DEBUG__
  Serial.print("C");
#endif

  bool passedChecksum = false;
  int count = 0;

  while (!passedChecksum) {
    count += 1;
	if (getMessage(controllerLength, controllerPin)) {
		passedChecksum = calculateChecksum(controllerLength);
	}
	else {
		// Serial.println(" exit with no controller message");
		return;
	}
  }

#ifdef __DEBUG__
  printBuffer(controllerLength);
#endif

  heat = bytebuffer[4];
  air = bytebuffer[0];

#ifdef __WARN__
  if (count > 1) {
    Serial.print("[!] WARN: Took ");
    Serial.print(count);
    Serial.println(" tries to read controller.");
  }
#endif
}


int filtWeight = 80; // Artisan default
void filtTemp(double v) {

	if (abs(temp - v) > 10)
		temp = v;
	else
		temp = (v * (100.0f - filtWeight) + temp * (float)filtWeight) / 100.0f;

} // end filtTemp


void getRoasterMessage() {
#ifdef __DEBUG__
  Serial.print("R ");
#endif

  bool passedChecksum = false;
  int count = 0;

  while (!passedChecksum && count < 3) {
    count += 1;
	if (getMessage(roasterLength, roasterPin)) {
		passedChecksum = calculateChecksum(roasterLength);
  }
	else {
		//Serial.println(" exit with no roaster message");
		temp = 1;
		return;
	}

  }
#ifdef __DEBUG__
  Serial.print("roaster buffer: ");
  printBuffer(roasterLength);
#endif

#ifdef __WARN__
  if (count > 1) {
    Serial.print("[!] WARN: Took ");
    Serial.print(count);
    Serial.println(" tries to read roaster.");
  }
#endif

  // temp = calculateTemp();
  filtTemp(calculateTemp());

}

void handleREAD() {
	Serial.print(25.0);
	Serial.print(',');
	Serial.print(0.0);
	Serial.print(',');
	Serial.print(temp);
	Serial.print(',');
	Serial.print(heat);
	Serial.print(',');
	Serial.print(air);
	Serial.print(',');
	// instead SV send raw temperature data, for diagnostic purposes
	if (temp <= 110) 
		Serial.println(inty);
	else
		Serial.println(intx);

}


void loop() {
  toggle = !toggle;
  unsigned long start = millis();
  if (toggle == 0) {
	  //Serial.print(millis()); Serial.println(">");
	  getControllerMessage();
	  //Serial.print("="); Serial.println(millis() - start);
  } else {
	 //Serial.print(millis()); Serial.println("<");
	 getRoasterMessage();
	 // Serial.print("="); Serial.println(millis() - start);
  }
  // Serial.println("");

    while (Serial.available() > 0) {
		String command = Serial.readStringUntil('\n');

		command.toUpperCase();

		if (command.startsWith("READ")) {
			handleREAD();
		}
		else if (command.startsWith("CHAN")) {
			Serial.println("# Active channels set to ...");
		}
		else if (command.startsWith("UNITS")) {
			CorF = command.charAt(6);
			Serial.print("# Changed units to "); Serial.println(CorF);
		}
		else if (command.startsWith("FILT")) {
			// FILT,lev1,lev2,lev3,lev4
			filtWeight = command.substring(5, 7).toInt();
			Serial.print(F("# Physical channel 1 filter set to ")); Serial.println(filtWeight);
		}

    } // end while serial available
  //}
}
